###############################################################################################
# Useful functions to assign the appropiate genotype and phenotyp means among closest probands for relatives (who only have phenotypes)
# Yanbing Wang (yanbing@bu.edu)
################################################################################################

# Required Ipute files
# 1. Phenotype file for probands: need to contain the columns for Y_proband, famID and ID
all_phenotype_probands<-read.csv("proband_pheno.csv",stringsAsFactors=FALSE)
# 2. Phenotype file for probands: need to contain the columns for Y_relative, famID and ID
all_phenotype_relatives<-read.csv("relative_pheno.csv",stringsAsFactors=FALSE)
# 3. The kinship matrix matrix, which will used to define the closest proband(s) for relatives
all_kinship2<-read.csv("all_kinship2.csv",stringsAsFactors=FALSE)
rownames(all_kinship2)<-colnames(all_kinship2)
# 2. Genotype file for probands: need to contain the columns of variants, and ID
all_genotype_probands<-read.csv("proband_genotype.csv",stringsAsFactors=FALSE)


################################################################################################
# --Define the closest proband(s) based on kinship coefficient for all relatives across families

allfamily<-unique(all_phenotype_probands$famID) #returns the vector of famID


# manage_relative() function will return the list, which contains:
# 1) ID(s) for corresponding closest proband(s) who have the largest kinship coefficient for each relative;
# 2) Corresponding kinship coefficient
manage_relative<-function(i_fam,all_phenotype_probands,all_phenotype_relatives){
  i_fam<-i_fam
  fam_index<-which(unlist( lapply( strsplit( colnames(all_kinship2),'_'),function(x) paste(x[2]))) == paste("fam",i_fam,sep=""))
  kin<-all_kinship2[fam_index,fam_index] #relationship (kinship) matrix for each single family
  
  proband_index<-which(colnames(kin)%in% all_phenotype_probands$ID)
  relative_index<-which(colnames(kin)%in% all_phenotype_relatives$ID)
  sub_kin<-as.matrix(kin[relative_index,proband_index]) 
  keep<-apply(sub_kin,1, function(x) sum(x)>0) #if the relative is not related with any genotypes, then remove
  sub_kin<-as.matrix(sub_kin[keep,])
  
  #mange results:
  sub_proband_index<-which(colnames(sub_kin)%in% all_phenotype_probands$ID)
  sub_relative_index<-which(rownames(sub_kin)%in% all_phenotype_relatives$ID)
  sub_relative_id<-rownames(sub_kin)[sub_relative_index]
  result<-vector("list",length = length(sub_relative_id))
  names(result)<- sub_relative_id
  
  i<-1
  for (ith_relative in  sub_relative_index){
    result[[i]]=c(rownames(as.matrix(sub_kin[ith_relative,]))[sub_kin[ith_relative,]==max(sub_kin[ith_relative,])],max(sub_kin[ith_relative,]))
    i<-i+1 
  }
  return(result)
}


# Loop to generate the information lists for relatives for each family
for (i in 1:length(allfamily)){
  code<-paste("fam",i,"_relatives<-manage_relative(",i,",all_phenotype_probands,all_phenotype_relatives)",sep = "")
  eval(parse(text=code))
}  


################################################################################################
# --Assign genotype means from closest probands for relatives

assign_genotypes<-function(upperMAF=0.01){
  
relatives_genotype<-function(N,proband_genotype){
  eval(parse(text = paste("famN_relatives<-fam",N,"_relatives",sep = "")))
  relative_genotype<-matrix(nrow = length(names(famN_relatives)),ncol=ncol(proband_genotype))
  relative_info<-matrix(nrow = length(names(famN_relatives)),ncol=2)
  rownames(relative_genotype)<-names(famN_relatives)
  colnames(relative_genotype)<-colnames(proband_genotype)
  rownames(relative_info)<-names(famN_relatives)
  colnames(relative_info)<-c("probandID","kin_coef")
  
  for (j in 1:length(famN_relatives)) {
    last<-length(famN_relatives[[j]])
    select_proband<-famN_relatives[[j]][-c(last)]#probandID will use genotype from this proband
    if(length(select_proband)==1){
      relative_genotype[j,]<-as.matrix(proband_genotype[which(rownames(proband_genotype) %in% select_proband),])
      #relative_info[j,]<-famN_relatives[[j]][c((last-2):last)]
    }else{
      relative_genotype[j,]<-colMeans(proband_genotype[which(rownames(proband_genotype) %in% select_proband),])}
    #relative_info[j,]<-famN_relatives[[j]][c((last-2):last)]
  }
  relative_genotype<-as.data.frame(relative_genotype)
  out<-cbind(relative_genotype)
  return(out)
}

##proband_genotype only contains genotypes for probands, with the rowname as the proband ID
proband_genotype<-all_genotype_probands[,-which(names(all_genotype_probands) %in% "ID")]
maf1_proband<-colSums(proband_genotype)/(2*nrow(proband_genotype))
mac_proband<-colSums(proband_genotype)
proband_genotype<-proband_genotype[,which( mac_proband>1 & maf1_proband< upperMAF)] #set the MAF upper bound
rownames(proband_genotype)<-all_genotype_probands$ID

##assign genotype mean from proband for relatives
for (i in allfamily) {
  eval(parse(text = paste("famN_relatives<-fam",i,"_relatives",sep = "")))
  if(length(famN_relatives)!=0){
    eval(parse(text = paste("relative_genotype_",i,"<-relatives_genotype(",i,",proband_genotype)",sep = "")))
  }
}

#combine the genotypes for all relatives
relative_genotype_file<-paste("relative_genotype_",allfamily,sep = "")
datalist<-lapply(relative_genotype_file, function(x){eval(parse(text = x))})
relative_genotype<-Reduce(function(x,y) {rbind(x,y)}, datalist)
all_relative_genotype<-relative_genotype
all_relative_genotype$ID<-rownames(relative_genotype)

return(all_relative_genotype)
}

################################################################################################
###########Assign phenotypes mean from closest probands for relatives, which will use as an covariate in the relative model

assign_phenotype<-function(){
  
relatives_phenotype<-function(N){
  eval(parse(text = paste("famN_relatives<-fam",N,"_relatives",sep = "")))
  relative_phenotype<-matrix(nrow = length(names(famN_relatives)),ncol=1)
  relative_info<-matrix(nrow = length(names(famN_relatives)),ncol=1)
  rownames(relative_phenotype)<-names(famN_relatives)
  colnames(relative_phenotype)<-"Ymean_proband"
  rownames(relative_info)<-names(famN_relatives)
  colnames(relative_info)<-c("kin_coef")
  
  for (j in 1:length(famN_relatives)) {
    last<-length(famN_relatives[[j]])
    select_proband<-famN_relatives[[j]][-c(last)]
    if(length(select_proband)==1){
      relative_phenotype[j,]<-all_phenotype_probands$Y_proband[which(all_phenotype_probands$ID %in% select_proband)]
      relative_info[j,]<-famN_relatives[[j]][c(last)]
    }else{
      relative_phenotype[j,]<-mean(all_phenotype_probands$Y_proband[which(all_phenotype_probands$ID %in% select_proband)])
      relative_info[j,]<-famN_relatives[[j]][c(last)]
    }
  }
  relative_phenotype<-as.data.frame(relative_phenotype)
  out<-cbind(relative_phenotype,relative_info)
  
  return(out)
}

# create relative phenotype for each family
for (i in allfamily) {
  eval(parse(text = paste("famN_relatives<-fam",i,"_relatives",sep = "")))
  if(length(famN_relatives)!=0){
    eval(parse(text = paste("relative_phenotype_",i,"<-relatives_phenotype(",i,")",sep = "")))
  }
}

relative_phenotype_file<-paste("relative_phenotype_",allfamily,sep = "")
datalist<-lapply(relative_phenotype_file, function(x){eval(parse(text = x))})
relative_phenotype<-Reduce(function(x,y) {rbind(x,y)}, datalist)
all_relative_phenotype1<-relative_phenotype
all_relative_phenotype1$ID<-rownames(relative_phenotype)
all_relative_phenotype<-merge(all_phenotype_relatives,all_relative_phenotype1,by="ID")
all_relative_phenotype<-all_relative_phenotype[order(all_relative_phenotype$famID),]
return(all_relative_phenotype)
}



