#########################################################################
# family-based Family History Aggregation-based Test (FHAT) and Optimal FHAT (FHAT-O)
# Yanbing Wang (yanbing@bu.edu)
#########################################################################

# Definition
# probands: who have available phenotypes+genotypes
# relatives: who have available phenotypes

library(CompQuadForm) # required package; Davies method for p-value calculation
wuweights <- function(maf) ifelse(maf > 0, dbeta(maf, 1, 25), 0) # default weights is WU's weights: beta density with parameters 1 and 25

library(GMMAT) #we use glmmkin to fit the GLMM
source("assign_phenotype_genotype.R") #contains useful functions to assign the appropiate genotype and phenotyp means among closest probands for relatives

# Required Ipute files
# 1. Phenotype file for probands: need to contain the columns of "Y_proband"= phenotype of probands, "famID"= family ID, "ID"= sample ID
## test data: all_phenotype_probands<-read.csv("/restricted/projectnb/yanbing-thesis/famFHATO/Package Functions/proband_pheno.csv",stringsAsFactors=FALSE)
# 2. Phenotype file for probands: need to contain the columns for "Y_relative"= phenotype of relatives, "famID"= family ID, "ID"= sample ID
## test data: all_phenotype_relatives<-read.csv("/restricted/projectnb/yanbing-thesis/famFHATO/Package Functions/relative_pheno.csv",stringsAsFactors=FALSE)
# 3. The kinship matrix matrix with the colname/rowname= sampe ID, which will used to define the closest proband(s) for relatives 
## test data: all_kinship2<-read.csv("/restricted/projectnb/yanbing-thesis/famFHATO/Package Functions/all_kinship2.csv",stringsAsFactors=FALSE)
## test data: rownames(all_kinship2)<-colnames(all_kinship2)
# 4. Genotype file for probands: need to contain the columns for variants, and "ID"= sample ID
## test data: all_genotype_probands<-read.csv("/restricted/projectnb/yanbing-thesis/famFHATO/Package Functions/proband_genotype.csv",stringsAsFactors=FALSE)

# Create genotypes for relatives, with a MAF upper bound
#all_genotype_relatives<-assign_genotypes(upperMAF = 0.01)

# Add column "Ymean_proband"= phenotype means across the closest probands, and column "kin_coef"= corresponding kinship coefficient
#all_phenotype_relatives<-assign_phenotype()


# FUNCTION: The null GLMM models for probands and relatives
# FUNCTION ARGUMENTS
# data_proband: Phenotype data of proband.
# data_relative: Phenotype data of relative. The data should contain a column for corresponding probands' disease status. If the relative is
#correlated to multiple probands, we take the mean of disease status among the closest probands (based on kinship coefficient).
# Y_proband: the outcome variable in probands
# Y_relative: the outcome variable in relatives
# covariates_proband: a list of covariates to be adjusted in the proband model
# covariates_relative: a list of covariates to be adjusted in the relative model. Note that the covariate should include probands' disease status as the covariate.



# FUNCTION: To perform FHAT or FHAT-O analysis
# FUNCTION ARGUMENTS
# proband_pheno: phenotype data of proband. The data should contain a column "ID" to indicate probands' IDs
# proband_geno: genotypes in probands. The data should contain a column "ID" to indicate probands' ID
# relative_pheno: phenotype data of relative, which can be obtained using functions in assign_phenotype_genotype.R. The data should contain a column "ID" to indicate relative' IDs, and the column for appropriate probands' phenotypes that will be used as one of the covariate
# relative_geno: genotypes in probands, which can be obtained using functions in assign_phenotype_genotype.R, where we assign the genotype means among closest proband(s) for each relative. The data should contain a column "ID" to indicate relatives' ID
# null_model_p: the null GLMM for probands
# null_model_p: the null GLMM for relatives
# upperMAF: the threshold of maximum minor allele frequencies(MAF) in analysis

main_function<-function(proband_pheno,proband_geno,relative_pheno,relative_geno,null_model_p,null_model_r,weights_func = wuweights,upperMAF=0.01){
# merge data
proband_data<-merge(all_phenotype_probands,all_genotype_probands,by="ID") #match by ID
relative_data<-merge(all_phenotype_relatives,all_genotype_relatives,by="ID") #match by ID
n_sample_p<-nrow(proband_data)
n_sample_r<-nrow(relative_data)

# genotype data for probands and relatives
use_proband_genotype<-as.matrix(proband_data[,-which(names(proband_data)%in% names(all_phenotype_probands))])
maf1_proband<-colSums(use_proband_genotype)/(2*nrow(use_proband_genotype))
mac_proband<-colSums(use_proband_genotype)
use_proband_genotype<-use_proband_genotype[,which( mac_proband>1 & maf1_proband< upperMAF)]#probands' genotypes

use_relative_genotype1<-as.matrix(relative_data[,-which(names(relative_data)%in% names(all_phenotype_relatives))]) #relatives' genotypes

# down-weight the genotypes for relatives by corresponding kinship coefficient
relative_data$kin_coef<-as.numeric(as.character(relative_data$kin_coef))
use_relative_genotype<-as.matrix(use_relative_genotype1*relative_data$kin_coef*2) #downweight

# scores
# note: can use SKAT:::SKATBinary_Robust() to correct Q score for robust adjustment
maf_p<-colSums(use_proband_genotype)/(2*n_sample_p)
weights_p <- wuweights(maf_p)
Z_p <- t( t( use_proband_genotype) * weights_p)

maf_r<-colSums(use_relative_genotype)/(2*n_sample_r)
weights_r <- wuweights(maf_r)
Z_r <- t( t( use_relative_genotype) * weights_r)

Q.tprob <- t(model_p$residuals)  %*% Z_p
Q.tr <- t(model_r$residuals)  %*% Z_r
B.r<-sum(model_r$residuals*(use_relative_genotype %*% weights_r))

#probandQ<-Q.tprob %*% t(Q.tprob)#proband-famSKAT Q
probandB<-sum(model_p$residuals*(use_proband_genotype %*% weights_p))

Q.tcom <- Q.tprob + Q.tr #combine scores

combineQ <- Q.tcom %*% t(Q.tcom)# famFHAT Q
combineB <- (probandB+B.r)^2

# lambdas
X1_p<-model_p$X
X1_r<-model_r$X


if(class(try(t(Z_p)%*% model_p$P  %*% Z_p))=="try-error"){
  W.1p<-t(Z_p)%*%model_p$Sigma_i%*%Z_p-t(Z_p)%*%model_p$Sigma_iX%*%solve(t(X1_p)%*%model_p$Sigma_iX)%*%t(model_p$Sigma_iX)%*%Z_p
}else{W.1p<-t(Z_p)%*% model_p$P  %*% Z_p}


if(class(try(t(Z_r)%*% model_r$P  %*% Z_r))=="try-error"){
  W.1r<-t(Z_r)%*%model_r$Sigma_i%*%Z_r-t(Z_r)%*%model_r$Sigma_iX%*%solve(t(X1_r)%*%model_r$Sigma_iX)%*%t(model_r$Sigma_iX)%*%Z_r
}else{W.1r<-t(Z_r)%*% model_r$P  %*% Z_r}

#lambda_p<-eigen(W.1p, symmetric = T, only.values = T)$values

W.com<-W.1p+W.1r#combine designed matrix 
lambda_com<-eigen((W.com),symmetric=TRUE, only.values = TRUE)$values #famFHAT eigenvalues


famFHATP<-davies(combineQ, lambda_com,lim = 10000,acc=10^(-6), rep(1,length(lambda_com)))$Qq# p-value

# famFHATo: calculate P value
r.all<-seq(0,1,0.1)
n.r<-length(r.all)
Q.all<-t(as.matrix(c(combineQ) * (rep(1,length(r.all))-r.all)+c(combineB)*r.all))
Phi<-W.com
p.m<-dim(Phi)[2]
lambda.all<-list()
for(i in 1:n.r){
  r.corr<-r.all[i]
  R.M<-diag(rep(1-r.corr,p.m)) + matrix(rep(r.corr,p.m*p.m),ncol=p.m)
  L<-chol(R.M,pivot=TRUE)
  options(warn = -1)
  K1<-L %*% (Phi %*% t(L))
  lambda.all[[i]]<-eigen((K1),symmetric=TRUE, only.values = TRUE)$values
}

famFHATOP<-SKAT:::SKAT_META_Optimal_Get_Pvalue(Q.all, W.com, r.all, method="optimal")$p.value
return(data.frame(famFHAT_pvalue=famFHATP,famFHATO_pvalue=famFHATOP))
}




# EXAMPLE  

# Required Ipute files
# 1. Phenotype file for probands: need to contain the columns of "Y_proband"= phenotype of probands, "famID"= family ID, "ID"= sample ID
all_phenotype_probands<-read.csv("proband_pheno.csv",stringsAsFactors=FALSE)
# 2. Phenotype file for probands: need to contain the columns for "Y_relative"= phenotype of relatives, "famID"= family ID, "ID"= sample ID
all_phenotype_relatives<-read.csv("relative_pheno.csv",stringsAsFactors=FALSE)
# 3. The kinship matrix matrix with the colname/rowname= sampe ID, which will used to define the closest proband(s) for relatives 
all_kinship2<-read.csv("all_kinship2.csv",stringsAsFactors=FALSE) #can be created using library(kinship2)
rownames(all_kinship2)<-colnames(all_kinship2)
# 2. Genotype file for probands: need to contain the columns for variants, and "ID"= sample ID
all_genotype_probands<-read.csv("proband_genotype.csv",stringsAsFactors=FALSE)


# --Create genotypes for relatives, with a MAF upper bound
all_genotype_relatives<-assign_genotypes(upperMAF = 0.01)

# --Add column "Ymean_proband"= phenotype means across the closest probands, and column "kin_coef"= corresponding kinship coefficient
all_phenotype_relatives<-assign_phenotype()

# --Fit the GLMM using glmmkin() from GMMAT
# relationship matrix: can use either the theoretical kinship matrix based on the known pedigree structures, or the empirical kinship matrix estimated from KING
all_kinship2<-as.matrix(all_kinship2)
# proband model
model_p <- GMMAT::glmmkin( Y_proband ~ proband_sex + proband_age ,id= "ID" ,family = binomial(link="logit") ,kins = all_kinship2,data= all_phenotype_probands)
# relative model: note the covariate should include probands' phenotype (we use the phenotype mean among closest probands)
model_r <- GMMAT::glmmkin( Y_relative ~ relative_sex + relative_age + Ymean_proband ,id= "ID" ,family = binomial(link="logit") ,kins = all_kinship2, data= all_phenotype_relatives)

# --Run analysis
main_function(proband_pheno=all_phenotype_probands,proband_geno= all_genotype_probands,
              relative_pheno=all_phenotype_relatives,relative_geno=all_genotype_relatives,
              null_model_p=model_p,null_model_r=model_r,weights_func = wuweights,upperMAF=0.01)

